; File:       ConvertKX.inc
; Written by: RuiLoureiro
;-------------------------------------------------------------------------
ConvertReal10KX  proto  :DWORD,:DWORD

$MagicLog2_32    equ 1292913987     ; = log(2) * 2^32
$MagicLog2_16    equ 19728          ; = log(2)  * 2^16
$MagicLog10_16   equ 217705         ; = log2(10) * 2^16

    dp_1    equ 17      ; = decimal places minus 1
    dp4     equ dp_1+4 
; 
; Input:
;           esi = 10
;
EXPONENTREAL_10K MACRO

                mov     bx, '+e'

                or      eax, eax
                jns     short @F
                
                mov     bx, '-e'
                neg     eax                     ; make number positive
        @@:      
                mov     word ptr [edi], bx

                ;EAX = 0 to 9999

                mov     edx,4294968          ;2^32/1000
                ;mov     ebx,10               ;per digit
                mul     edx                  ;extend first digit
                mov     ecx,edx              ;digit 1 in CL
                ;mul     ebx                  ;second digit
                mul     esi
                mov     ch,dl                ;digit 2 in CH
                ;mul     ebx                  ;third digit
                mul     esi
                bswap   ecx                  ;digits 1 & 2 up high
                mov     ch,dl                ;digit 3 in CH
                ;mul     ebx                  ;digit 4 in DL
                mul     esi
                lea     ecx,[edx+ecx+'0000'] ;combine and make ASCII
                bswap   ecx                  ;re-order bytes
                ; ----------------------
                ; ECX = "0000" to "9999" 
                ; ----------------------
                mov     [edi+2], ecx
                add     edi, 6                    
ENDM
; 
;                               Data segment
; 
;---------------------------------------------
$maxexp     equ 90
$maxcounter equ 2*$maxexp+1
;---------------------------------------------
.data 
_realexp4931    dt 1.0e+4931

;---------------------------------------------
    cntr = -$maxexp
    REPEAT $maxcounter
        IF cntr NE 0
            dt @CatStr(<1.0e>,%-cntr)
        ELSE
 _PowerTable12  dt 1.0
        ENDIF
            dw 0
        cntr = cntr + 1
    ENDM
; 
; 
;                               code segment
; 
comment * 
REAL10
  Exponent                     Significand                  Meaning

    0000       0000_0000_0000_0000                          Signed Zero
               0000_0000_0000_0001 to 7FFF_FFFF_FFFF_FFFF   Signed Denormal
               8000_0000_0000_0000 to FFFF_FFFF_FFFF_FFFF   Signed Pseudo-Denormal

0001 to 7FFE   0000_0000_0000_0000 to 7FFF_FFFF_FFFF_FFFF   Invalid
               8000_0000_0000_0000 to FFFF_FFFF_FFFF_FFFF   Signed Normal

    7FFF       0000_0000_0000_0000 to 7FFF_FFFF_FFFF_FFFF   Invalid
               8000_0000_0000_0000                          Signed Infinity
               8000_0000_0000_0001 to BFFF_FFFF_FFFF_FFFF   Signaling NaN
               C000_0000_0000_0000                          Indefinite Quiet NaN
               C000_0000_0000_0001 to FFFF_FFFF_FFFF_FFFF   Quiet NaN
               
           *
; 
.code
        _shifttableX   	dd _shift0X    ; +0 
                        dd _shift1X    ; +4
                        dd _shift2X    ; +8
                        dd _shift3X    ;+12
                        dd _shift4X    ;+16

ConvertReal10KX         proc     pReal10:DWORD, pStr:DWORD
                        LOCAL    iExp       :DWORD
                        LOCAL    aExp       :DWORD
                        LOCAL    realx[3]   :DWORD
                        LOCAL    temp1      :DWORD
                        LOCAL    temp2      :DWORD
                        ;
                        push     ebx
                        push     esi
                        push     edi

                        ;finit

	                  mov     edi, pStr
                        mov     ecx, pReal10
	                  fld     tbyte ptr [ecx]

	                  mov     aExp, 0

                        ; ---------------------------
                        ;        remove sign
                        ;    move real10 to realx
                        ; if iExp=0 we use this realx
                        ; ---------------------------
	                  movzx   edx, word ptr [ecx+8]

                        mov     al, ' '
                        test    edx, 8000h
                        jz      short @F
                        mov     al, '-'
                @@:     mov     [edi], al

	                  and     edx, 7FFFh           ; remove sign

	                  mov     ebx, [ecx+0]
	                  mov     eax, [ecx+4]

	                  mov     [realx+0], ebx
	                  mov     [realx+4], eax
	                  mov     [realx+8], edx       ; exponent
	                  jnz     short _expnotzero               

                        ; ------------------------
                        ;       exponent = 0
                        ; ------------------------
                        or      ebx, eax
                        jz      _iszero
                        
                        ;-------------------------
                        ;    It is Denormal or
                        ;     Pseudo-Denormal
                        ;-------------------------
                        fld     _realexp4931     ; st(0)=1.0e+4931
	                  mov     aExp,-4931
	                  fmul                     ; normalization
	                  fld     st               ; get a copy
	                  fstp    tbyte ptr realx
                        ;
	                  mov     ebx, [realx+0]
	                  mov     eax, [realx+4]
	                  mov     edx, [realx+8]

                        ; --------------------
                        ;     remove sign
                        ; --------------------
	                  and     edx, 7FFFh           ; remove sign
	                  mov     [realx+8], edx
	                  jmp     short @F

        _expnotzero:    cmp     edx, 7FFFh
	                  je      _isinfinity
                        ;
                        ;
                        ;
                        and     eax, 80000000h
                        jz      _erro1
                        
                        ;---------------------
                        ;       start
                        ;---------------------
                @@:     mov     esi, 10
                                                    
                        ;----------------------------------
                        ; iExp= 4 * Exp because we multiply
                        ;  by 2^16 and divide by 2^14 only
                        ;----------------------------------
	                  sub     edx, 3FFFh           ; =16383
	                  imul    edx, $MagicLog2_16   ; log(2)*2^16
	                  sar     edx, 16              ; divide by 2^16
	                  mov     iExp, edx
                        jz      short _getrealx
                        
	                  cmp     edx, $maxexp
	                  jg      a1
	                  cmp     edx,-$maxexp
	                  jge     short _load
;------------------------------------------------
                 a1:    neg     edx 	
	                  mov     temp1, edx           

                        ; ---------------------
                        ;       temp2=
                        ; (log2(10)*2^16)/2^16
                        ; ---------------------
	                  imul    edx, $MagicLog10_16          ; = 217705
	                  sar     edx, 16
	                  mov     temp2, edx

	                  fild    temp2       ; get the characteristic
	                  fild    temp1       ; log2(10)*expscale
                        ;
	                  fldl2t		  ; log2(10)
                        fmul                ; log2(10)*expscale
                            
                        fsub    st,st(1) ; get only the mantissa but keep the characteristic
                        f2xm1                   ; 2^(mantissa)-1
                        fld1
                        fadd                    ; add 1 and pop
                        fscale
                        fstp    st(1)           ; remove the characteristic
                        jmp     short @F        ;_multiply
                        
;--------------------------------------------------------        
            _load:	sal     edx, 2                       ; edx = edx*4 
                        fld     [_PowerTable12+edx+edx*2]    ; edx = edx*12 
;--------------------------------------------------------                

                @@:     fmul                    ; X * 10^expscaleS = d.??????
             
	                  fld      st              ; get a copy 
	                  fstp     tbyte ptr realx	
	                  xor      edx, edx

                        ; ----------------------------
                        ;  fraction bits in EAX:EBX
                        ; ----------------------------
        _getrealx:      mov      ecx, [realx+8]
	                  mov      ebx, [realx+0]
                        mov      eax, [realx+4]
                        ;
	                  and      ecx, 7FFFh
	                  sub      ecx, 16382      ; 10=> exponent = 16386

	                  cmp      ecx, 4
	                  ja       _shift0X
	                  jmp      [_shifttableX+ecx*4]

           _shift4X::   test     eax, 60000000h              ; 6= 0110B
	                  jz       short a6

                        ; ------------------------
                        ; correct exponent * 10^-1
                        ; exponent = exponent + 1
                        ; ------------------------
	                  fld      [_PowerTable12+12]          ; = 0.1
	                  fmul 
	                  add      iExp, 1

	                  fld      st                  ; NEW get a copy 
	                  fstp     tbyte ptr realx

	                  mov      eax, [realx+4]
                        mov      ebx, [realx+0]

                        ; -------------------------------
                        ;    fraction bits in EAX:EBX
                        ;    shift: EDX <- EAX <- EBX 
                        ; -------------------------------                            
	                  shld     edx, eax,1
	                  shld     eax, ebx,1
	                  shl      ebx, 1
	                  add      ebx, 4
	                  jmp      _shift0X

                a6:	shld     edx, eax, 4        ;ecx=1 ebx+4
	                  shld     eax, ebx, 4        ;ecx=2 ebx+8
	                  shl      ebx, 4             ;ecx=3 ebx+9
	                  add      ebx, 12            ;ecx=4 ebx+12
	                  jmp      _shift0X

                        ; -------------------------------
                        ;    fraction bits in EAX:EBX
                        ;    shift: EDX <- EAX <- EBX 
                        ; -------------------------------
           _shift3X::   shld     edx, eax, 3
	                  shld     eax, ebx, 3
	                  shl      ebx, 3              ; * 8
	                  add      ebx, 9    ;10
	                  jmp      _shift0X

                        ; -------------------------------
                        ;    fraction bits in EAX:EBX
                        ;    shift: EDX <- EAX <- EBX 
                        ; -------------------------------
           _shift2X::   shld     edx, eax, 2
	                  shld     eax, ebx, 2
	                  shl      ebx, 2              ; * 4
	                  add      ebx, 8
	                  jmp      _shift0X

                        ; -------------------------------
                        ;    fraction bits in EAX:EBX
                        ;    shift: EDX <- EAX <- EBX 
                        ; -------------------------------
           _shift1X::   shld     edx, eax, 1
	                  shld     eax, ebx, 1
	                  shl      ebx, 1              ; * 2
	                  add      ebx, 4
;-------------------------------------------------------
                        ; -------------------------------
                        ;  fraction bits in EDX:EAX:EBX
                        ; -------------------------------
           _shift0X::   adc      eax, 0             ; add carry to eax and edx
	                  adc      edx, 0
                        ;
	                  mov      [edi+1], dl
	                  mov      byte ptr [edi+2],'.'
	                  or       byte ptr [edi+1],'0'
                        ;------------------------------------------------
                        ;            bit f1 is bit 31 in eax
                        ;
                        ;                 eax          ebx
                        ;                               10 (first by eax)
                        ;      ----------------------------
                        ;         edx     ecx
                        ;                 edx          eax
                        ;    ++++++++++++++++++++++++++++++
                        ;   edx+carry     ecx          ebx
                        ;------------------------------------------------
                    	k=3
                        REPEAT dp_1
                            mul     esi             ; edx:eax=eax*10
                            mov     [edi+k], dl     ; save DL
                            mov     ecx, eax        ; save EAX in ecx
                            ;
                            mov     eax, ebx
                            mul     esi             ; ebx=ebx*10
                            add     ecx, edx
                            adc     byte ptr [edi+k],'0'
                            ;
                            mov     ebx, eax        ; new ebx
                            mov     eax, ecx        ; new eax
                    	k=k+1	
                        ENDM

	                  mul      esi
	                  mov      [edi+k], dl
	                  mov      ecx, eax            ; save ecx
                        ;
	                  mov      eax, ebx
	                  mul      esi
	                  add      ecx, edx
	                  adc      byte ptr [edi+k],'0'

                        add      edi, dp4

;---------------------------------------------------------
	                  mov      eax, iExp
	                  add      eax, aExp
                        jz       short _finish 

                        EXPONENTREAL_10K

            _finish:    mov      eax, edi
                        mov      edi, pStr
                        sub      eax, edi            

            _exit0:     ;mov      dword ptr [edi-4], eax
                        mov      byte ptr [edi+eax], 0

                        fstp     st(0)                      ; remove realx
                                              
                        xor      eax, eax
                        clc
            _exit:      pop      edi
                        pop      esi
                        pop      ebx
                        ret
;-----------------------------------------------------------
    _isinfinity:        or       eax, ebx
                        cmp      eax, 80000000h
                        jne      short _erro1               ; is NAN
                        
                        mov      dword ptr [edi+1], "IFNI"
                        mov      dword ptr [edi+5], "YTIN"
                        mov      eax, 9

                        jmp      _exit0
                   
                        ;--------------------------
                        ; value to be converted = 0
                        ;--------------------------
        _iszero:        mov      word ptr [edi], 3020h
                        mov      eax, 2
                        jmp      _exit0
                    
                        ; --------------------
                        ; is indefinite or NAN
                        ; --------------------
        _erro1:         mov      eax, 1
        
                        ; -------------------
                        ; -------------------
        _error:         fstp     st(0)                      ; remove realx
        
                        mov      dword ptr [edi+0], "ORRE"
                        mov      byte ptr [edi+4], "R"
                        ;
                        mov      dword ptr [edi-4], 5
                        mov      byte ptr [edi+5], 0
                        stc
                        jmp      _exit                                        
ConvertReal10KX         endp
; 
